extern crate tracing;

use alloc::format;
use nom::branch::alt;
use nom::bytes::complete::take_until;
use nom::character::complete::one_of;
use nom::combinator::rest;
use nom::multi::fold_many1;
use nom::sequence::{pair, preceded};
use time::format_description::well_known::iso8601::{
    EncodedConfig, OffsetPrecision, TimePrecision,
};

use core::hash::BuildHasher;

use crate::{Error, IntersperseSource};
use alloc::borrow::Cow;
use alloc::string::String;
use core::convert::{From, TryInto};
use core::fmt::{Debug, Display};

use time::format_description::well_known::iso8601::{Config as TimeConfig, Iso8601};
use time::{Duration, OffsetDateTime};

#[derive(Clone, Debug, PartialEq, Eq)]
pub struct AckParsed {
    pub a_type: icinga_api::api::acknowledge_problem_action::TargetType,
    pub filter: String,
    pub filter_vars: FilterVars,
    pub comment: String,
}

#[derive(Clone, Debug, PartialEq, Eq)]
pub struct AckExpParsed {
    pub a_type: icinga_api::api::acknowledge_problem_action::TargetType,
    pub filter: String,
    pub filter_vars: FilterVars,
    pub end_time: OffsetDateTime,
    pub comment: String,
}

#[derive(Clone, Debug, PartialEq, Eq)]
pub struct DtParsed {
    pub a_type: icinga_api::api::schedule_downtime_action::TargetType,
    pub filter: String,
    pub filter_vars: FilterVars,
    pub end_time: OffsetDateTime,
    pub comment: String,
}

#[derive(Clone, Debug, PartialEq, Eq)]
pub struct DtAtParsed {
    pub a_type: icinga_api::api::schedule_downtime_action::TargetType,
    pub filter: String,
    pub filter_vars: FilterVars,
    pub start_time: OffsetDateTime,
    pub end_time: OffsetDateTime,
    pub comment: String,
}

#[derive(Clone, Debug, PartialEq, Eq)]
pub struct ComParsed {
    pub a_type: icinga_api::api::add_comment_action::TargetType,
    pub filter: String,
    pub filter_vars: FilterVars,
    pub comment: String,
}

#[derive(Clone, Debug, PartialEq, Eq)]
pub struct RmParsed {
    pub a_type: icinga_api::api::object_type::ObjectType,
    pub filter: String,
    pub filter_vars: FilterVars,
}

/// Parsed command
#[derive(Clone, Debug, PartialEq, Eq)]
pub enum ParseResult {
    /// Acknowledge a problem (`ack <host>[!<service>[!<hoster>]] <comment>`)
    Ack(AckParsed),
    /// Temporarily acknowledge a problem (`ackexp <host>[!<service>[!<hoster>]] <end_time> <comment>`)
    AckExp(AckExpParsed),
    /// Schedule a downtime (`dt <host>[!<service>[!<hoster>]] <end_time> <comment>`)
    Dt(DtParsed),
    /// Schedule a downtime (`dtat <host>[!<service>[!<hoster>]] <start_time> <end_time> <comment>`)
    DtAt(DtAtParsed),
    /// Add a comment to an object (`com <host>[!<service>[!<hoster>]] <comment>`)
    Com(ComParsed),
    /// Remove all ack and downtime on an object (`rm <host>[!<service>[!<hoster>]]`)
    Rm(RmParsed),
    /// Activate or deactivate notifications (`notif <true|false>`)
    Notif(bool),
    /// Add oneself to those who should be notified of new alerts (`hl`)
    Highlight,
    /// Show help message (`help`)
    Help,
    Unknown,
}

#[must_use]
#[derive(Clone, Copy, Debug, PartialEq, Eq, Hash)]
pub enum Parser {
    Ack,
    AckExp,
    Dt,
    DtAt,
    Com,
    Rm,
    Notif,
    Highlight,
    Help,
    Ping,
    Unknown,
}

impl Parser {
    /// Dispatch parsing to related functions.
    ///
    /// # Errors
    ///
    /// This function will return an error if parsing fail.
    #[inline]
    pub fn parse<'a, I, S>(&self, arguments: I) -> Result<ParseResult, ParseActionError>
    where
        I: IntoIterator<Item = &'a S> + Debug,
        S: AsRef<str> + 'a + ?Sized,
        <I as IntoIterator>::IntoIter: Debug,
    {
        let res = match self {
            Parser::Ack => parse_ack(arguments, false),
            Parser::AckExp => parse_ack(arguments, true),
            Parser::Dt => parse_dt(arguments),
            Parser::DtAt => parse_dtat(arguments),
            Parser::Com => parse_com(arguments),
            Parser::Rm => parse_rm(arguments),
            Parser::Notif => parse_notif(
                arguments
                    .into_iter()
                    .next()
                    .map(AsRef::as_ref)
                    .unwrap_or_default(),
            ),
            Parser::Highlight => Ok(ParseResult::Highlight),
            Parser::Help => Ok(ParseResult::Help),
            Parser::Ping => {
                tracing::info!("pong");
                Ok(ParseResult::Unknown)
            }
            Parser::Unknown => Ok(ParseResult::Unknown),
        };
        if let Err(err) = &res {
            tracing::error!("Parsing failed with error: {err}");
        }
        res
    }
}

#[derive(Debug, Copy, Clone, Eq, PartialEq, Default, Hash)]
pub struct NoEndTimeError {}

impl Display for NoEndTimeError {
    fn fmt(&self, f: &mut core::fmt::Formatter<'_>) -> core::fmt::Result {
        write!(f, "No end time was found!")
    }
}

impl Error for NoEndTimeError {}

#[derive(Debug, Clone, Eq, PartialEq)]
pub enum ParseEndTimeError {
    NoEndTime(NoEndTimeError),
    Parsing(nom::error::ErrorKind),
}

impl Display for ParseEndTimeError {
    fn fmt(&self, f: &mut core::fmt::Formatter<'_>) -> core::fmt::Result {
        write!(f, "Couldn't parse end time!")
    }
}

impl From<nom::error::Error<&str>> for ParseEndTimeError {
    fn from(err: nom::error::Error<&str>) -> Self {
        Self::Parsing(err.code)
    }
}

impl From<NoEndTimeError> for ParseEndTimeError {
    fn from(err: NoEndTimeError) -> Self {
        Self::NoEndTime(err)
    }
}

impl Error for ParseEndTimeError {
    fn source(&self) -> Option<&(dyn Error + 'static)> {
        match self {
            Self::Parsing(_) => None,
            Self::NoEndTime(err) => Some(err),
        }
    }
}

#[derive(Debug, PartialEq)]
pub enum ParseActionError {
    GenericParse,
    MissingTarget,
    TargetParse(nom::Err<nom::error::Error<Box<str>>>),
    IncorrectStartTime(Option<time::error::Parse>),
    ParseEndTime(ParseEndTimeError),
}

impl Display for ParseActionError {
    fn fmt(&self, f: &mut core::fmt::Formatter<'_>) -> core::fmt::Result {
        match self {
            Self::GenericParse => f.write_str("Parsing error"),
            Self::MissingTarget => f.write_str("Target is missing from command"),
            Self::TargetParse(err) => write!(f, "Parsing of target failed with error : {err}"),
            Self::IncorrectStartTime(_) => f.write_str("The start time provided is incorrect"),
            Self::ParseEndTime(error) => Display::fmt(error, f),
        }
    }
}

impl From<ParseEndTimeError> for ParseActionError {
    fn from(err: ParseEndTimeError) -> Self {
        Self::ParseEndTime(err)
    }
}

impl Error for ParseActionError {
    fn source(&self) -> Option<&(dyn Error + 'static)> {
        match self {
            Self::IncorrectStartTime(None) | Self::GenericParse | Self::MissingTarget => None,
            Self::TargetParse(err) => Some(err),
            Self::IncorrectStartTime(Some(err)) => Some(err),
            Self::ParseEndTime(err) => Some(err),
        }
    }
}

impl ParseActionError {
    fn from_target_parse(err: nom::Err<nom::error::Error<&str>>) -> Self {
        let err = match err {
            nom::Err::Error(err) => {
                let err = nom::error::Error::new(Box::from(err.input), err.code);
                nom::Err::Error(err)
            }
            nom::Err::Failure(err) => {
                let err = nom::error::Error::new(Box::from(err.input), err.code);
                nom::Err::Failure(err)
            }
            nom::Err::Incomplete(needed) => nom::Err::Incomplete(needed),
        };
        Self::TargetParse(err)
    }
}

/// Parse the duration of an action
///
/// # Error
///
/// Return a [`ParseEndTimeError`](self::ParseEndTimeError) in case of any problem in the parsing:
///
/// * [`ParseInt`](self::ParseEndTimeError::ParseInt) if the internal parsing of numbers fail
/// * [`ParseChar`](self::ParseEndTimeError::ParseChar) if the internal parsing of char fail
///   (shouldn't ever happen)
/// * [`NoEndTime`](self::ParseEndTimeError::NoEndTime) if there is nothing to parse
///
#[tracing::instrument(level = "trace")]
pub fn parse_end_time(unp_time: &str) -> Result<Duration, ParseEndTimeError> {
    const PATTERN: &[char] = &['w', 'd', 'h', 'm', 's'];

    fn to_sec(i: i64, c: char) -> i64 {
        match c {
            'w' => i.saturating_mul(604_800_i64),
            'd' => i.saturating_mul(86_400_i64),
            'h' => i.saturating_mul(3_600_i64),
            'm' => i.saturating_mul(60_i64),
            's' => i,
            _ => unreachable!(),
        }
    }

    // Check if what we have might be an end time and directly get out otherwise.
    if unp_time.is_empty() || !unp_time.chars().next().unwrap().is_numeric() {
        return Err(NoEndTimeError::default().into());
    }

    let (res, time) = fold_many1(
        pair(
            nom::character::complete::i64::<&str, nom::error::Error<&str>>,
            one_of(PATTERN),
        ),
        || 0_i64,
        |acc, (i, c)| acc.saturating_add(to_sec(i, c)),
    )(unp_time)
    .map_err(|err| match err {
        nom::Err::Error(err) | nom::Err::Failure(err) => match err.code {
            nom::error::ErrorKind::Many1 => ParseEndTimeError::NoEndTime(NoEndTimeError::default()),
            _ => ParseEndTimeError::from(err),
        },
        nom::Err::Incomplete(_) => unreachable!(),
    })?;

    if !res.is_empty() {
        return Err(NoEndTimeError::default().into());
    }

    Ok(Duration::new(time, 0))
}

#[derive(Debug, PartialEq)]
pub struct Target {
    pub a_type: icinga_api::api::object_type::ObjectType,
    pub filter: String,
    pub filter_vars: FilterVars,
}

#[derive(Debug, Clone, PartialEq, Eq, Default)]
pub struct FilterVars {
    pub host: String,
    pub service: Option<String>,
    pub hoster: Option<String>,
}

impl<S: Default + BuildHasher> From<FilterVars> for crate::HashMap<String, String, S> {
    fn from(value: FilterVars) -> Self {
        let s = S::default();
        let mut ret = Self::with_capacity_and_hasher(2, s);
        ret.insert(String::from("HostName"), value.host);
        if let Some(service) = value.service {
            ret.insert(String::from("ServiceName"), service);
        }
        if let Some(hoster) = value.hoster {
            ret.insert(String::from("Hoster"), hoster);
        }
        ret
    }
}

#[tracing::instrument(level = "trace")]
fn parse_target(target: &str) -> core::result::Result<Target, nom::Err<nom::error::Error<&str>>> {
    let a_type: icinga_api::api::object_type::ObjectType;
    let filter: &str;

    let filter_vars: FilterVars;

    let (target, host) = take_until::<_, _, nom::error::Error<&str>>("!")(target)
        .map_or((None, target), |(res, host)| (Some(res), host));
    let host = format!("^{host}$");

    if let Some(target) = target {
        let (target, service) = preceded(
            nom::character::complete::char('!'),
            alt((take_until("!"), rest)),
        )(target)?;

        let service = Some(format!("^{service}$"));
        a_type = icinga_api::api::object_type::ObjectType::Service;

        if target.is_empty() {
            filter_vars = FilterVars { host, service, hoster: None };
            filter = "regex(HostName, host.name) && regex(ServiceName, service.name)";
        } else {
            let (_, hoster) = preceded(nom::character::complete::char('!'), rest)(target)?;

            let hoster = Some(format!("^{hoster}$"));
            filter_vars = FilterVars { host, service, hoster };
            filter = "regex(HostName, host.name) && regex(ServiceName, service.name) && regex(Hoster, host.vars.computer_hoster)";
        }
    } else {
        a_type = icinga_api::api::object_type::ObjectType::Host;
        filter_vars = FilterVars { host, service: None, hoster: None };
        filter = "regex(HostName, host.name)";
    }

    Ok(Target {
        a_type,
        filter: String::from(filter),
        filter_vars,
    })
}

#[tracing::instrument(level = "trace")]
pub fn parse_comment<'a, I, S>(comment: I) -> String
where
    I: IntoIterator<Item = &'a S> + Debug,
    S: AsRef<str> + 'a + ?Sized,
{
    fn map_comment(word: &str) -> Cow<str> {
        let (suffix, ticket) = nom::combinator::opt(preceded(
            nom::character::complete::char::<_, nom::error::Error<&str>>('#'),
            nom::character::complete::digit1,
        ))(word)
        .unwrap();
        if let Some(ticket) = ticket {
            format!("[#{ticket}](https://bugs.evolix.org/{ticket}){suffix}").into()
        } else {
            Cow::Borrowed(word)
        }
    }

    IntersperseSource::intersperse(
        comment
            .into_iter()
            .map(AsRef::as_ref)
            .flat_map(str::split_whitespace)
            .map(map_comment),
        Cow::Borrowed(" "),
    )
    .collect()
}

/// This function parse a given command and provide the [`Dispatcher`] to the correct parser.
#[inline]
pub fn parse_command(command: &str) -> Parser {
    match command {
        "ack" => Parser::Ack,
        "ackexp" => Parser::AckExp,
        "dt" => Parser::Dt,
        "dtat" => Parser::DtAt,
        "com" => Parser::Com,
        "rm" => Parser::Rm,
        "ping" => Parser::Ping,
        "notif" => Parser::Notif,
        "hl" => Parser::Highlight,
        "help" => Parser::Help,
        _ => Parser::Unknown,
    }
}

#[cfg(all(not(test), feature = "std"))]
fn add_now_to_duration(duration: Duration) -> OffsetDateTime {
    OffsetDateTime::now_utc().saturating_add(duration)
}

#[cfg(any(test, not(feature = "std")))]
const fn add_now_to_duration(duration: Duration) -> OffsetDateTime {
    OffsetDateTime::UNIX_EPOCH.saturating_add(duration)
}

/// Parse the .ack (`expire` = `false`) and .ackexp (`expire` = `true`) command
#[tracing::instrument(level = "debug")]
pub fn parse_ack<'a, I, S>(arguments: I, expire: bool) -> Result<ParseResult, ParseActionError>
where
    I: IntoIterator<Item = &'a S> + Debug,
    S: AsRef<str> + 'a + ?Sized,
    <I as IntoIterator>::IntoIter: Debug,
{
    let mut arguments = arguments.into_iter();
    let target = if let Some(target) = arguments.next() {
        parse_target(target.as_ref()).map_err(ParseActionError::from_target_parse)?
    } else {
        return Err(ParseActionError::MissingTarget);
    };

    let end_time = expire
        .then(|| {
            arguments
                .next()
                .map(|s| parse_end_time(s.as_ref()).map(add_now_to_duration))
                .transpose()?
                .ok_or_else(|| ParseEndTimeError::NoEndTime(NoEndTimeError::default()))
        })
        .transpose()?;

    let comment = parse_comment(arguments);

    let a_type = target.a_type.try_into().unwrap();
    let ret = if expire {
        ParseResult::AckExp(AckExpParsed {
            a_type,
            filter: target.filter,
            filter_vars: target.filter_vars,
            comment,
            end_time: end_time.unwrap(),
        })
    } else {
        ParseResult::Ack(AckParsed {
            a_type,
            filter: target.filter,
            filter_vars: target.filter_vars,
            comment,
        })
    };
    Ok(ret)
}

#[tracing::instrument(level = "debug")]
pub fn parse_dt<'a, I, S>(arguments: I) -> Result<ParseResult, ParseActionError>
where
    I: IntoIterator<Item = &'a S> + Debug,
    S: AsRef<str> + 'a + ?Sized,
    <I as IntoIterator>::IntoIter: Debug,
{
    let mut arguments = arguments.into_iter();
    let target = if let Some(target) = arguments.next() {
        parse_target(target.as_ref()).map_err(ParseActionError::from_target_parse)?
    } else {
        return Err(ParseActionError::MissingTarget);
    };

    let end_time = arguments
        .next()
        .map(|s| parse_end_time(s.as_ref()).map(add_now_to_duration))
        .transpose()?
        .ok_or_else(|| ParseEndTimeError::NoEndTime(NoEndTimeError::default()))?;

    let comment = parse_comment(arguments);

    Ok(ParseResult::Dt(DtParsed {
        a_type: target.a_type.try_into().unwrap(),
        filter: target.filter,
        filter_vars: target.filter_vars,
        comment,
        end_time,
    }))
}

#[tracing::instrument(level = "debug")]
pub fn parse_dtat<'a, I, S>(arguments: I) -> Result<ParseResult, ParseActionError>
where
    I: IntoIterator<Item = &'a S> + Debug,
    S: AsRef<str> + 'a + ?Sized,
    <I as IntoIterator>::IntoIter: Debug,
{
    fn parse_start_time(start_time: &str) -> Result<OffsetDateTime, ParseActionError> {
        const PARSE_CONFIG: EncodedConfig = TimeConfig::DEFAULT
            .set_time_precision(TimePrecision::Minute {
                decimal_digits: None,
            })
            .set_offset_precision(OffsetPrecision::Minute)
            .encode();
        const PARSABLE: Iso8601<PARSE_CONFIG> = Iso8601::<PARSE_CONFIG>;

        OffsetDateTime::parse(start_time, &PARSABLE)
            .map_err(|e| ParseActionError::IncorrectStartTime(Some(e)))
    }

    let mut arguments = arguments.into_iter();
    let target = if let Some(target) = arguments.next() {
        parse_target(target.as_ref()).map_err(ParseActionError::from_target_parse)?
    } else {
        return Err(ParseActionError::MissingTarget);
    };

    let start_time = parse_start_time(
        arguments
            .next()
            .map(AsRef::as_ref)
            .ok_or(ParseActionError::IncorrectStartTime(None))?,
    )?;

    let end_time = arguments
        .next()
        .map(|s| parse_end_time(s.as_ref()).map(|duration| start_time.saturating_add(duration)))
        .transpose()?
        .ok_or_else(|| ParseEndTimeError::NoEndTime(NoEndTimeError::default()))?;

    let comment = parse_comment(arguments);

    Ok(ParseResult::DtAt(DtAtParsed {
        a_type: target.a_type.try_into().unwrap(),
        filter: target.filter,
        filter_vars: target.filter_vars,
        comment,
        start_time,
        end_time,
    }))
}

#[tracing::instrument(level = "debug")]
pub fn parse_com<'a, I, S>(arguments: I) -> Result<ParseResult, ParseActionError>
where
    I: IntoIterator<Item = &'a S> + Debug,
    S: AsRef<str> + 'a + ?Sized,
    <I as IntoIterator>::IntoIter: Debug,
{
    let mut arguments = arguments.into_iter();
    let target = if let Some(target) = arguments.next() {
        parse_target(target.as_ref()).map_err(ParseActionError::from_target_parse)?
    } else {
        return Err(ParseActionError::MissingTarget);
    };

    let comment = parse_comment(arguments);

    Ok(ParseResult::Com(ComParsed {
        a_type: target.a_type.try_into().unwrap(),
        filter: target.filter,
        filter_vars: target.filter_vars,
        comment,
    }))
}

/// Parse argument for notif command.
///
/// # Errors
///
/// This function will return an error if the argument isn't a boolean.
#[inline]
pub fn parse_notif(argument: &str) -> Result<ParseResult, ParseActionError> {
    argument
        .parse::<bool>()
        .map(ParseResult::Notif)
        .map_err(|_| ParseActionError::GenericParse)
}

#[tracing::instrument(level = "debug")]
pub fn parse_rm<'a, I, S>(arguments: I) -> Result<ParseResult, ParseActionError>
where
    I: IntoIterator<Item = &'a S> + Debug,
    S: AsRef<str> + 'a + ?Sized,
{
    let target = if let Some(target) = arguments.into_iter().next() {
        parse_target(target.as_ref()).map_err(ParseActionError::from_target_parse)?
    } else {
        return Err(ParseActionError::MissingTarget);
    };

    Ok(ParseResult::Rm(RmParsed {
        a_type: target.a_type,
        filter: target.filter,
        filter_vars: target.filter_vars,
    }))
}

#[cfg(test)]
mod test {
    extern crate test;
    use test::black_box;

    use super::{Duration, OffsetDateTime, String};

    mod end_time {
        use super::super::{parse_end_time, ParseEndTimeError};
        use super::{black_box, test, Duration};

        #[test]
        fn simple() {
            assert_eq!(parse_end_time("1s"), Ok(Duration::SECOND));
            assert_eq!(parse_end_time("1m"), Ok(Duration::MINUTE));
            assert_eq!(parse_end_time("1h"), Ok(Duration::HOUR));
            assert_eq!(parse_end_time("1d"), Ok(Duration::DAY));
            assert_eq!(parse_end_time("1w"), Ok(Duration::WEEK));
        }

        #[test]
        fn complexe() {
            const EXPECTED: Duration = Duration::weeks(5)
                .saturating_add(Duration::minutes(20))
                .saturating_add(Duration::seconds(6));

            assert_eq!(parse_end_time("5w20m6s"), Ok(EXPECTED));
            assert_eq!(parse_end_time("20m5w6s"), Ok(EXPECTED));
            assert_eq!(parse_end_time("6s20m5w"), Ok(EXPECTED));
        }

        #[test]
        fn none() {
            let result = parse_end_time("none").unwrap_err();
            match result {
                ParseEndTimeError::NoEndTime(_) => (),
                ParseEndTimeError::Parsing(_) => panic!("result should be a NoEndTimeError !"),
            };
        }

        #[test]
        fn no_end_time() {
            let error = parse_end_time("no end time").unwrap_err();
            match error {
                ParseEndTimeError::NoEndTime(_) => (),
                ParseEndTimeError::Parsing(_) => panic!("result should be a NoEndTimeError !"),
            }
        }

        #[test]
        fn no_suffix() {
            assert!(parse_end_time("42").is_err());
            assert!(parse_end_time("42h13").is_err());
        }

        #[bench]
        fn bench(b: &mut test::Bencher) {
            b.iter(|| {
                let to_parse = black_box("5w20m6s");
                black_box(parse_end_time(to_parse))
            });
        }
    }

    mod target {
        use crate::parsing::FilterVars;

        use super::super::{parse_target, Target};
        use super::{black_box, test, String};

        #[test]
        fn host() {
            let expected_vars = FilterVars{ host: String::from("^test$"), ..FilterVars::default() };

            let expected = Target {
                a_type: icinga_api::api::object_type::ObjectType::Host,
                filter: String::from("regex(HostName, host.name)"),
                filter_vars: expected_vars,
            };

            assert_eq!(parse_target("test"), Ok(expected));
        }

        #[test]
        fn service() {
            let expected_vars = FilterVars{
                host: String::from("^test_host$"),
                service: Some(String::from("^test_service$")),
                ..FilterVars::default()
            };

            let expected = Target {
                a_type: icinga_api::api::object_type::ObjectType::Service,
                filter: String::from(
                    "regex(HostName, host.name) && regex(ServiceName, service.name)",
                ),
                filter_vars: expected_vars,
            };

            assert_eq!(parse_target("test_host!test_service"), Ok(expected));
        }

        #[test]
        fn service_with_hoster() {
            let expected_vars = FilterVars{
                host: String::from("^test_host$"),
                service: Some(String::from("^test_service$")),
                hoster: Some(String::from("^test_hoster$")),
            };

            let expected = Target {
                a_type: icinga_api::api::object_type::ObjectType::Service,
                filter: String::from("regex(HostName, host.name) && regex(ServiceName, service.name) && regex(Hoster, host.vars.computer_hoster)"),
                filter_vars: expected_vars,
            };

            assert_eq!(
                parse_target("test_host!test_service!test_hoster"),
                Ok(expected)
            );
        }

        #[bench]
        fn bench_host(b: &mut test::Bencher) {
            b.iter(|| {
                let target = black_box("test_host");
                black_box(parse_target(target))
            });
        }

        #[bench]
        fn bench_service(b: &mut test::Bencher) {
            b.iter(|| {
                let target = black_box("test_host!test_service");
                black_box(parse_target(target))
            });
        }

        #[bench]
        fn bench_service_with_hoster(b: &mut test::Bencher) {
            b.iter(|| {
                let target = black_box("test_host!test_service!test_hoster");
                black_box(parse_target(target))
            });
        }
    }

    mod comment {
        use super::{black_box, test};

        use super::super::parse_comment;

        #[test]
        fn simple() {
            const LOREM_IPSUM: &str = "Rerum vel vero doloremque eligendi cum. Nobis exercitationem voluptatem blanditiis et delectus sequi quibusdam. Accusamus assumenda iusto aspernatur. Rerum quos voluptatem incidunt est. Non consectetur quo est. Cum dolorem optio ipsum sapiente cumque voluptatibus laudantium doloribus.";

            assert_eq!(
                parse_comment(LOREM_IPSUM.split_ascii_whitespace()),
                LOREM_IPSUM
            );
        }

        #[test]
        fn with_ticket_simple() {
            const COMMENT: &str = "Ticket #12345 : test";
            const COMMENT2: &str = "#12345";

            assert_eq!(
                parse_comment(COMMENT.split_ascii_whitespace()),
                "Ticket [#12345](https://bugs.evolix.org/12345) : test"
            );

            assert_eq!(
                parse_comment(COMMENT2.split_ascii_whitespace()),
                "[#12345](https://bugs.evolix.org/12345)"
            );
        }

        #[test]
        fn with_ticket() {
            const COMMENT: &str = "Ticket #12345: test";

            assert_eq!(
                parse_comment(COMMENT.split_ascii_whitespace()),
                "Ticket [#12345](https://bugs.evolix.org/12345): test"
            );
        }

        #[bench]
        fn bench_simple(b: &mut test::Bencher) {
            // While we could use str::split_ascii_whitespace instead, the objective is to bench parse_comment() so might as well do this work ourselves.
            const LOREM_IPSUM: [&str; 6] =
                ["Rerum", "vel", "vero", "doloremque", "eligendi", "cum."];

            b.iter(|| {
                let lorem_ipsum = black_box(LOREM_IPSUM);
                black_box(parse_comment(lorem_ipsum))
            });
        }

        #[bench]
        fn bench_ticket(b: &mut test::Bencher) {
            // While we could use str::split_ascii_whitespace instead, the objective is to bench parse_comment() so might as well do this work ourselves.
            const COMMENT: [&str; 5] = ["Ticket", "#12345", ":", "test", "comment"];

            b.iter(|| {
                let comment = black_box(COMMENT);
                black_box(parse_comment(comment))
            });
        }
    }

    mod command {
        use time::macros::datetime;

        use super::super::{parse_command, ParseResult, Parser};
        use super::{black_box, test, Duration, OffsetDateTime};

        #[test]
        fn ack() {
            const COMMAND: &str = "ack test This is a test";
            let (command, dispatcher) = {
                let mut command = COMMAND.split_ascii_whitespace();
                let dispatcher = parse_command(command.next().unwrap_or_default());
                (command, dispatcher)
            };

            assert_eq!(dispatcher, Parser::Ack);

            let ParseResult::Ack(result) = dispatcher.parse(command).unwrap() else { unreachable!() };
            assert_eq!(
                result.a_type,
                icinga_api::api::acknowledge_problem_action::TargetType::Host
            );
            assert_eq!(result.comment, "This is a test");
            assert_eq!(result.filter, "regex(HostName, host.name)");
            assert_eq!(
                result
                    .filter_vars
                    .host,
                "^test$"
            );
        }

        #[test]
        fn ackexp() {
            const COMMAND: &str = "ackexp test 1w2d3h6m12s This is a test";
            const EXPECTED_DURATION: Duration = Duration::weeks(1)
                .saturating_add(Duration::days(2))
                .saturating_add(Duration::hours(3))
                .saturating_add(Duration::minutes(6))
                .saturating_add(Duration::seconds(12));
            const EXPECTED_END_TIME: OffsetDateTime =
                OffsetDateTime::UNIX_EPOCH.saturating_add(EXPECTED_DURATION);

            let (command, dispatcher) = {
                let mut command = COMMAND.split_ascii_whitespace();
                let dispatcher = parse_command(command.next().unwrap_or_default());
                (command, dispatcher)
            };

            assert_eq!(dispatcher, Parser::AckExp);

            let ParseResult::AckExp(result) = dispatcher.parse(command).unwrap() else { unreachable!() };
            assert_eq!(
                result.a_type,
                icinga_api::api::acknowledge_problem_action::TargetType::Host
            );
            assert_eq!(result.comment, "This is a test");
            assert_eq!(result.filter, "regex(HostName, host.name)");
            assert_eq!(
                result
                    .filter_vars
                    .host,
                "^test$"
            );
            assert_eq!(result.end_time, EXPECTED_END_TIME);
        }

        #[test]
        fn dt() {
            const COMMAND: &str = "dt test 1w2d4h8m16s This is a test";
            const EXPECTED_DURATION: Duration = Duration::weeks(1)
                .saturating_add(Duration::days(2))
                .saturating_add(Duration::hours(4))
                .saturating_add(Duration::minutes(8))
                .saturating_add(Duration::seconds(16));
            const EXPECTED_END_TIME: OffsetDateTime =
                OffsetDateTime::UNIX_EPOCH.saturating_add(EXPECTED_DURATION);

            let (command, dispatcher) = {
                let mut command = COMMAND.split_ascii_whitespace();
                let dispatcher = parse_command(command.next().unwrap_or_default());
                (command, dispatcher)
            };

            assert_eq!(dispatcher, Parser::Dt);

            let ParseResult::Dt(result) = dispatcher.parse(command).unwrap() else { unreachable!() };
            assert_eq!(
                result.a_type,
                icinga_api::api::schedule_downtime_action::TargetType::Host
            );
            assert_eq!(result.comment, "This is a test");
            assert_eq!(result.filter, "regex(HostName, host.name)");
            assert_eq!(
                result
                    .filter_vars
                    .host,
                "^test$"
            );
            assert_eq!(result.end_time, EXPECTED_END_TIME);
        }

        #[test]
        fn dtat() {
            const COMMAND: &str = "dtat test 2020-01-01T00:00-01:00 1w2d4h8m16s This is a test";
            const EXPECTED_DURATION: Duration = Duration::weeks(1)
                .saturating_add(Duration::days(2))
                .saturating_add(Duration::hours(4))
                .saturating_add(Duration::minutes(8))
                .saturating_add(Duration::seconds(16));
            const EXPECTED_START_TIME: OffsetDateTime = datetime!(2020-01-01 0:00 -1);
            const EXPECTED_END_TIME: OffsetDateTime =
                EXPECTED_START_TIME.saturating_add(EXPECTED_DURATION);

            let (command, dispatcher) = {
                let mut command = COMMAND.split_ascii_whitespace();
                let dispatcher = parse_command(command.next().unwrap_or_default());
                (command, dispatcher)
            };

            assert_eq!(dispatcher, Parser::DtAt);

            let ParseResult::DtAt(result) = dispatcher.parse(command).unwrap() else { unreachable!() };
            assert_eq!(
                result.a_type,
                icinga_api::api::schedule_downtime_action::TargetType::Host
            );
            assert_eq!(result.comment, "This is a test");
            assert_eq!(result.filter, "regex(HostName, host.name)");
            assert_eq!(
                result
                    .filter_vars
                    .host,
                "^test$"
            );
            assert_eq!(result.end_time, EXPECTED_END_TIME);
            assert_eq!(result.start_time, EXPECTED_START_TIME);
        }

        #[test]
        fn com() {
            const COMMAND: &str = "com test This is a test";
            let (command, dispatcher) = {
                let mut command = COMMAND.split_ascii_whitespace();
                let dispatcher = parse_command(command.next().unwrap_or_default());
                (command, dispatcher)
            };

            assert_eq!(dispatcher, Parser::Com);
            let ParseResult::Com(result) = dispatcher.parse(command).unwrap() else { unreachable!() };
            assert_eq!(
                result.a_type,
                icinga_api::api::add_comment_action::TargetType::Host
            );
            assert_eq!(result.comment, "This is a test");
            assert_eq!(result.filter, "regex(HostName, host.name)");
            assert_eq!(
                result
                    .filter_vars
                    .host,
                "^test$"
            );
        }

        #[test]
        fn rm() {
            const COMMAND: &str = "rm test";
            let (command, dispatcher) = {
                let mut command = COMMAND.split_ascii_whitespace();
                let dispatcher = parse_command(command.next().unwrap_or_default());
                (command, dispatcher)
            };

            assert_eq!(dispatcher, Parser::Rm);
            let ParseResult::Rm(result) = dispatcher.parse(command).unwrap() else { unreachable!() };
            assert_eq!(
                result.a_type,
                icinga_api::api::object_type::ObjectType::Host
            );
            assert_eq!(result.filter, "regex(HostName, host.name)");
            assert_eq!(
                result
                    .filter_vars
                    .host,
                "^test$"
            );
        }

        #[test]
        fn notif() {
            const COMMAND_TRUE: &str = "notif true";
            const COMMAND_FALSE: &str = "notif false";

            let mut command_true = COMMAND_TRUE.split_ascii_whitespace();
            let result_true = parse_command(command_true.next().unwrap_or_default())
                .parse(command_true)
                .unwrap();

            assert_eq!(result_true, ParseResult::Notif(true));

            let mut command_false = COMMAND_FALSE.split_ascii_whitespace();
            let result_false = parse_command(command_false.next().unwrap_or_default())
                .parse(command_false)
                .unwrap();

            assert_eq!(result_false, ParseResult::Notif(false));
        }

        #[test]
        fn highlight() {
            const COMMAND: &str = "hl";
            let dispatcher = parse_command(COMMAND);

            assert_eq!(dispatcher, Parser::Highlight);
            assert_eq!(
                dispatcher.parse(core::iter::empty::<&str>()).unwrap(),
                ParseResult::Highlight
            );
        }

        #[test]
        fn help() {
            const COMMAND: &str = "help";
            let dispatcher = parse_command(COMMAND);

            assert_eq!(dispatcher, Parser::Help);
            assert_eq!(
                dispatcher.parse(core::iter::empty::<&str>()).unwrap(),
                ParseResult::Help
            );
        }

        #[bench]
        fn bench_ackexp(b: &mut test::Bencher) {
            const COMMAND: [&str; 7] = [
                "ackexp",
                "test_host!test_service",
                "1w2d4h8m16s",
                "This",
                "is",
                "a",
                "test.",
            ];

            b.iter(|| {
                let command = black_box(COMMAND);

                black_box(parse_command(command[0]).parse(&command[1..]))
            });
        }

        #[bench]
        fn bench_dt(b: &mut test::Bencher) {
            const COMMAND: [&str; 7] = [
                "dt",
                "test_host!test_service",
                "1w2d4h8m16s",
                "This",
                "is",
                "a",
                "test.",
            ];

            b.iter(|| {
                let command = black_box(COMMAND);

                black_box(parse_command(command[0]).parse(&command[1..]))
            });
        }

        #[bench]
        fn bench_dtat(b: &mut test::Bencher) {
            const COMMAND: [&str; 8] = [
                "dtat",
                "test_host!test_service",
                "2020-01-01T00:00-01:00",
                "1w2d4h8m16s",
                "This",
                "is",
                "a",
                "test.",
            ];

            b.iter(|| {
                let command = black_box(COMMAND);

                black_box(parse_command(command[0]).parse(&command[1..]))
            });
        }

        #[bench]
        fn bench_com(b: &mut test::Bencher) {
            const COMMAND: [&str; 6] =
                ["com", "test_host!test_service", "This", "is", "a", "test."];

            b.iter(|| {
                let command = black_box(COMMAND);

                black_box(parse_command(command[0]).parse(&command[1..]))
            });
        }

        #[bench]
        fn bench_rm(b: &mut test::Bencher) {
            const COMMAND: [&str; 2] = ["rm", "test_host!test_service"];

            b.iter(|| {
                let command = black_box(COMMAND);

                black_box(parse_command(command[0]).parse(&command[1..]))
            });
        }

        #[bench]
        fn bench_hl(b: &mut test::Bencher) {
            const COMMAND: [&str; 1] = ["hl"];

            b.iter(|| {
                let command = black_box(COMMAND);

                black_box(parse_command(command[0]).parse(core::iter::empty::<&str>()))
            });
        }

        #[bench]
        fn bench_help(b: &mut test::Bencher) {
            const COMMAND: [&str; 1] = ["help"];

            b.iter(|| {
                let command = black_box(COMMAND);

                black_box(parse_command(command[0]).parse(core::iter::empty::<&str>()))
            });
        }
    }
}
